@@ -1,20 +1,14 @@
-""" File-indexing functionality. """
+"""File-indexing functionality. """
 
 import os
-import re
 import json
-from keyword import iskeyword
-import warnings
-from copy import deepcopy
-from collections import defaultdict, namedtuple
-
+from collections import defaultdict
+from pathlib import Path
 from bids_validator import BIDSValidator
-
-from .writing import build_path, write_contents_to_file
-from .models import Config, BIDSFile, Entity, Tag, FileAssociation
+import bids.config
+from .models import Config, Entity, Tag, FileAssociation
 from ..utils import listify, make_bidsfile
-from ..config import get_option
-from ..external import six
+from ..exceptions import BIDSConflictingValuesError
 
 
 def _extract_entities(bidsfile, entities):
@@ -29,12 +23,12 @@ def _extract_entities(bidsfile, entities):
 
 
 def _check_path_matches_patterns(path, patterns):
-    ''' Check if the path matches at least one of the provided patterns. '''
+    """Check if the path matches at least one of the provided patterns. """
     if not patterns:
         return False
     path = os.path.abspath(path)
     for patt in patterns:
-        if isinstance(patt, six.string_types):
+        if isinstance(patt, str):
             if path == patt:
                 return True
         elif patt.search(path):
@@ -43,6 +37,11 @@ def _check_path_matches_patterns(path, patterns):
 
 
 class BIDSLayoutIndexer(object):
+    """ Indexer class for BIDSLayout.
+
+    Args:
+        layout (BIDSLayout): The BIDSLayout to index.
+    """
 
     def __init__(self, layout):
 
@@ -52,7 +51,6 @@ class BIDSLayoutIndexer(object):
         self.root = layout.root
         self.config_filename = layout.config_filename
         self.validator = BIDSValidator(index_associated=True)
-
         # Create copies of list attributes we'll modify during indexing
         self.config = list(layout.config.values())
         self.include_patterns = list(layout.force_index)
@@ -86,6 +84,7 @@ class BIDSLayoutIndexer(object):
         # to the BIDS project root.
         to_check = os.path.relpath(f, self.root)
         to_check = os.path.join(os.path.sep, to_check)
+        to_check = Path(to_check).as_posix()  # bids-validator works with posix paths only
         return self.validator.is_bids(to_check)
 
     def _index_dir(self, path, config, default_action=None):
@@ -97,7 +96,7 @@ class BIDSLayoutIndexer(object):
         if abs_path.startswith(os.path.join(self.root, 'derivatives')):
             return
 
-        config = list(config)     # Shallow copy
+        config = list(config)  # Shallow copy
 
         # Check for additional config file in directory
         layout_file = self.config_filename
@@ -134,11 +133,11 @@ class BIDSLayoutIndexer(object):
                 d = os.path.join(dirpath, d)
                 self._index_dir(d, list(config), default_action=default)
 
-            # prevent subdirectory traversal
+            # Prevent subdirectory traversal
             break
 
     def _index_file(self, f, dirpath, entities, default_action=None):
-        ''' Create DB record for file and its tags. '''
+        """Create DB record for file and its tags. """
         abs_fn = os.path.join(dirpath, f)
 
         # Skip files that fail validation, unless forcibly indexing
@@ -166,19 +165,54 @@ class BIDSLayoutIndexer(object):
         return bf
 
     def index_files(self):
-        ''' Index all files in the BIDS dataset. '''
+        """Index all files in the BIDS dataset. """
         self._index_dir(self.root, self.config)
 
-    def index_metadata(self):
-        ''' Index metadata for all files in the BIDS dataset. '''
+    def index_metadata(self, **filters):
+        """Index metadata for all files in the BIDS dataset.
+
+        Parameters
+        ----------
+
+        **filters
+            keyword arguments passed to the .get() method of a
+            :obj:`bids.layout.BIDSLayout` object.
+            These keyword arguments define what files get selected
+            for metadata indexing.
+        """
+
+        dot = '.' if bids.config.get_option('extension_initial_dot') else ''
+        
+        if filters:
+            # ensure we are returning objects
+            filters['return_type'] = 'object'
+            # until 0.11.0, user can specify extension or extensions
+            ext_key = 'extensions' if 'extensions' in filters else 'extension'
+            if filters.get(ext_key):
+                filters[ext_key] = listify(filters[ext_key])
+                # ensure json files are being indexed
+                # XXX 0.14: dot always == '.'
+                json_ext = dot + 'json'
+                if json_ext not in filters[ext_key]:
+                    filters[ext_key].append(json_ext)
+
         # Process JSON files first if we're indexing metadata
-        all_files = self.layout.get(absolute_paths=True)
+        all_files = self.layout.get(absolute_paths=True, **filters)
 
         # Track ALL entities we've seen in file names or metadatas
         all_entities = {}
         for c in self.config:
             all_entities.update(c.entities)
 
+        # If key/value pairs in JSON files duplicate ones extracted from files,
+        # we can end up with Tag collisions in the DB. To prevent this, we
+        # store all filename/entity pairs and the value, and then check against
+        # that before adding each new Tag.
+        all_tags = {}
+        for t in self.session.query(Tag).all():
+            key = '{}_{}'.format(t.file_path, t.entity_name)
+            all_tags[key] = str(t.value)
+
         # We build up a store of all file data as we iterate files. It looks
         # like: { extension/suffix: dirname: [(entities, payload)]}}.
         # The payload is left empty for non-JSON files.
@@ -194,9 +228,14 @@ class BIDSLayoutIndexer(object):
                 if key not in file_data:
                     file_data[key] = defaultdict(list)
 
-                if ext == 'json':
+                if ext == dot + 'json':
                     with open(bf.path, 'r') as handle:
-                        payload = json.load(handle)
+                        try:
+                            payload = json.load(handle)
+                        except json.JSONDecodeError as e:
+                            msg = ("Error occurred while trying to decode JSON"
+                                   " from file '{}'.".format(bf.path))
+                            raise IOError(msg) from e
                 else:
                     payload = None
 
@@ -238,7 +277,7 @@ class BIDSLayoutIndexer(object):
             # add the payload to the stack. Finally, we invert the
             # stack and merge the payloads in order.
             ext_key = "{}/{}".format(ext, suffix)
-            json_key = "json/{}".format(suffix)
+            json_key = dot + "json/{}".format(suffix)
             dirname = bf.dirname
 
             payloads = []
@@ -249,10 +288,10 @@ class BIDSLayoutIndexer(object):
                 json_data = file_data.get(json_key, {}).get(dirname, [])
                 for js_ents, js_md, js_path in json_data:
                     js_keys = set(js_ents.keys())
-                    if (js_keys - file_ent_keys):
+                    if js_keys - file_ent_keys:
                         continue
                     matches = [js_ents[name] == file_ents[name]
-                                for name in js_keys]
+                               for name in js_keys]
                     if all(matches):
                         payloads.append((js_md, js_path))
 
@@ -260,7 +299,7 @@ class BIDSLayoutIndexer(object):
                 candidates = file_data.get(ext_key, {}).get(dirname, [])
                 for ents, _, path in candidates:
                     keys = set(ents.keys())
-                    if (keys - file_ent_keys):
+                    if keys - file_ent_keys:
                         continue
                     matches = [ents[name] == file_ents[name] for name in keys]
                     if all(matches):
@@ -287,14 +326,14 @@ class BIDSLayoutIndexer(object):
             n_pl = len(payloads)
             for i, (pl, js_file) in enumerate(payloads):
                 if (i + 1) < n_pl:
-                    other = payloads[i+1][1]
+                    other = payloads[i + 1][1]
                     create_association_pair(js_file, other, 'Child', 'Parent')
 
             # Inheritance for current file
             n_pl = len(ancestors)
             for i, src in enumerate(ancestors):
                 if (i + 1) < n_pl:
-                    dst = ancestors[i+1]
+                    dst = ancestors[i + 1]
                     create_association_pair(src, dst, 'Child', 'Parent')
 
             # Files with IntendedFor field always get mapped to targets
@@ -309,7 +348,7 @@ class BIDSLayoutIndexer(object):
             # Link files to BOLD runs
             if suffix in ['physio', 'stim', 'events', 'sbref']:
                 images = self.layout.get(
-                    extension=['nii', 'nii.gz'], suffix='bold',
+                    extension=['.nii', '.nii.gz'], suffix='bold',
                     return_type='filename', **file_ents)
                 for img in images:
                     create_association_pair(bf.path, img, 'IntendedFor',
@@ -318,7 +357,7 @@ class BIDSLayoutIndexer(object):
             # Link files to DWI runs
             if suffix == 'sbref' or ext in ['bvec', 'bval']:
                 images = self.layout.get(
-                    extension=['nii', 'nii.gz'], suffix='dwi',
+                    extension=['.nii', '.nii.gz'], suffix='dwi',
                     return_type='filename', **file_ents)
                 for img in images:
                     create_association_pair(bf.path, img, 'IntendedFor',
@@ -326,6 +365,20 @@ class BIDSLayoutIndexer(object):
 
             # Create Tag <-> Entity mappings, and any newly discovered Entities
             for md_key, md_val in file_md.items():
+                tag_string = '{}_{}'.format(bf.path, md_key)
+                # Skip pairs that were already found in the filenames
+                if tag_string in all_tags:
+                    file_val = all_tags[tag_string]
+                    if str(md_val) != file_val:
+                        msg = (
+                            "Conflicting values found for entity '{}' in "
+                            "filename {} (value='{}') versus its JSON sidecar "
+                            "(value='{}'). Please reconcile this discrepancy."
+                        )
+                        raise BIDSConflictingValuesError(
+                            msg.format(md_key, bf.path, file_val,
+                            md_val))
+                    continue
                 if md_key not in all_entities:
                     all_entities[md_key] = Entity(md_key, is_metadata=True)
                     self.session.add(all_entities[md_key])
