@@ -1,10 +1,15 @@
+""" Tools for reading/writing BIDS data files. """
+
+from os.path import join
+import warnings
+import json
+
 import numpy as np
 import pandas as pd
-from os.path import join
+
 from bids.utils import listify
 from .entities import NodeIndex
 from .variables import SparseRunVariable, DenseRunVariable, SimpleVariable
-import warnings
 
 
 BASE_ENTITIES = ['subject', 'session', 'task', 'run']
@@ -13,38 +18,48 @@ ALL_ENTITIES = BASE_ENTITIES + ['datatype', 'suffix', 'acquisition']
 
 def load_variables(layout, types=None, levels=None, skip_empty=True,
                    dataset=None, scope='all', **kwargs):
-    ''' A convenience wrapper for one or more load_*_variables() calls.
+    """A convenience wrapper for one or more load_*_variables() calls.
 
-    Args:
-        layout (BIDSLayout): BIDSLayout containing variable files.
-        types (str, list): Types of variables to retrieve. All valid values
-            reflect the filename stipulated in the BIDS spec for each kind of
-            variable. Valid values include: 'events', 'physio', 'stim',
-            'scans', 'participants', 'sessions', and 'regressors'.
-        levels (str, list): Optional level(s) of variables to load. Valid
-            values are 'run', 'session', 'subject', or 'dataset'. This is
-            simply a shorthand way to specify types--e.g., 'run' will be
-            converted to types=['events', 'physio', 'stim', 'regressors'].
-        skip_empty (bool): Whether or not to skip empty Variables (i.e.,
-            where there are no rows/records in a file after applying any
-            filtering operations like dropping NaNs).
-        dataset (NodeIndex): An existing NodeIndex container to store the
-            loaded data in. Can be used to iteratively construct a dataset
-            that contains otherwise heterogeneous sets of variables. If None,
-            a new NodeIndex is used.
-        scope (str, list): The scope of the space to search for variables. See
-            docstring for BIDSLayout for details and valid predefined values.
-        kwargs: Optional keyword arguments to pass onto the individual
-            load_*_variables() calls.
+    Parameters
+    ----------
+    layout : :obj:`bids.layout.BIDSLayout`
+        BIDSLayout containing variable files.
+    types : str or list
+        Types of variables to retrieve. All valid values
+        reflect the filename stipulated in the BIDS spec for each kind of
+        variable. Valid values include: 'events', 'physio', 'stim',
+        'scans', 'participants', 'sessions', and 'regressors'.
+    levels : str or list
+        Optional level(s) of variables to load. Valid
+        values are 'run', 'session', 'subject', or 'dataset'. This is
+        simply a shorthand way to specify types--e.g., 'run' will be
+        converted to types=['events', 'physio', 'stim', 'regressors'].
+    skip_empty : bool
+        Whether or not to skip empty Variables (i.e.,
+        where there are no rows/records in a file after applying any
+        filtering operations like dropping NaNs).
+    dataset : NodeIndex
+        An existing NodeIndex container to store the
+        loaded data in. Can be used to iteratively construct a dataset
+        that contains otherwise heterogeneous sets of variables. If None,
+        a new NodeIndex is used.
+    scope : str or list
+        The scope of the space to search for variables. See
+        docstring for BIDSLayout for details and valid predefined values.
+    kwargs : dict
+        Optional keyword arguments to pass onto the individual
+        load_*_variables() calls.
 
-    Returns:
-        A NodeIndex instance.
+    Returns
+    -------
+    A NodeIndex instance.
 
-    Example:
-        >>> load_variables(layout, ['events', 'physio'], subject='01')
-        # returns all variables stored in _events.tsv and _physio.tsv.gz files
-        # for runs that belong to subject with id '01'.
-    '''
+    Examples
+    --------
+    >>> load_variables(layout, ['events', 'physio'], subject='01')  # doctest: +SKIP
+    # returns all variables stored in _events.tsv and _physio.tsv.gz files
+    # for runs that belong to subject with id '01'.
+    """
 
     TYPES = ['events', 'physio', 'stim', 'scans', 'participants', 'sessions',
              'regressors']
@@ -89,38 +104,51 @@ def _load_time_variables(layout, dataset=None, columns=None, scan_length=None,
                          drop_na=True, events=True, physio=True, stim=True,
                          regressors=True, skip_empty=True, scope='all',
                          **selectors):
-    ''' Loads all variables found in *_events.tsv files and returns them as a
+    """Loads all variables found in *_events.tsv files and returns them as a
     BIDSVariableCollection.
 
-    Args:
-        layout (BIDSLayout): A BIDSLayout to scan.
-        dataset (NodeIndex): A BIDS NodeIndex container. If None, a new one is
-            initialized.
-        columns (list): Optional list of names specifying which columns in the
-            event files to read. By default, reads all columns found.
-        scan_length (float): Optional duration of runs (in seconds). By
-            default, this will be extracted from the BOLD image. However, in
-            cases where the user doesn't have access to the images (e.g.,
-            because only file handles are locally available), a fixed duration
-            can be manually specified as a fallback.
-        drop_na (bool): If True, removes all events where amplitude is n/a. If
-            False, leaves n/a values intact. Note that in the latter case,
-            transformations that requires numeric values may fail.
-        events (bool): If True, extracts variables from events.tsv
-            files.
-        physio (bool): If True, extracts variables from _physio files.
-        stim (bool): If True, extracts variables from _stim files.
-        skip_empty (bool): Whether or not to skip empty Variables (i.e.,
-            where there are no rows/records in a file, or all onsets,
-            durations, and amplitudes are 0).
-        scope (str, list): The scope of the space to search for variables. See
-            docstring for BIDSLayout for details and valid predefined values.
-        selectors (dict): Optional keyword arguments passed onto the
-            BIDSLayout instance's get() method; can be used to constrain
-            which data are loaded.
+    Parameters
+    ----------
+    layout : :obj:`bids.layout.BIDSLayout`
+        A BIDSLayout to scan.
+    dataset : NodeIndex
+        A BIDS NodeIndex container. If None, a new one is
+        initialized.
+    columns : list
+        Optional list of names specifying which columns in the
+        event files to read. By default, reads all columns found.
+    scan_length : float
+        Optional duration of runs (in seconds). By
+        default, this will be extracted from the BOLD image. However, in
+        cases where the user doesn't have access to the images (e.g.,
+        because only file handles are locally available), a fixed duration
+        can be manually specified as a fallback.
+    drop_na : bool
+        If True, removes all events where amplitude is n/a. If
+        False, leaves n/a values intact. Note that in the latter case,
+        transformations that requires numeric values may fail.
+    events : bool
+        If True, extracts variables from events.tsv files.
+    physio : bool
+        If True, extracts variables from _physio files.
+    stim : bool
+        If True, extracts variables from _stim files.
+    skip_empty : bool
+        Whether or not to skip empty Variables (i.e.,
+        where there are no rows/records in a file, or all onsets,
+        durations, and amplitudes are 0).
+    scope : str or list
+        The scope of the space to search for variables. See
+        docstring for BIDSLayout for details and valid predefined values.
+    selectors : dict
+        Optional keyword arguments passed on to the
+        BIDSLayout instance's get() method; can be used to constrain
+        which data are loaded.
 
-    Returns: A NodeIndex instance.
-    '''
+    Returns
+    -------
+    A NodeIndex instance.
+    """
 
     # Extract any non-keyword arguments
     selectors = selectors.copy()
@@ -130,7 +158,7 @@ def _load_time_variables(layout, dataset=None, columns=None, scan_length=None,
 
     selectors['datatype'] = 'func'
     selectors['suffix'] = 'bold'
-    images = layout.get(return_type='object', extension='nii.gz',
+    images = layout.get(return_type='object', extension=['.nii', '.nii.gz'],
                         scope=scope, **selectors)
 
     if not images:
@@ -165,14 +193,48 @@ def _load_time_variables(layout, dataset=None, columns=None, scan_length=None,
                        "available, or manually specify the scan duration.")
                 raise ValueError(msg)
 
-        run = dataset.get_or_create_node('run', entities, image_file=img_f,
-                                         duration=duration, repetition_time=tr)
-        run_info = run.get_info()
+        # We don't want to pass all the image file's entities onto get_node(),
+        # as there can be unhashable nested slice timing values, and this also
+        # slows down querying unnecessarily. Instead, pick out files only based
+        # on the core BIDS entities and any entities explicitly passed as
+        # selectors.
+        # TODO: one downside of this approach is the stripped entities also
+        # won't be returned in the resulting node due to the way things are
+        # implemented. Consider adding a flag to control this.
+        select_on = {k: v for (k, v) in entities.items()
+                     if k in BASE_ENTITIES or k in selectors}
+
+        # If a matching node already exists, return it
+        result = dataset.get_nodes('run', select_on)
+
+        if result:
+            if len(result) > 1:
+                raise ValueError("More than one existing Node matches the "
+                                 "specified entities! You may need to pass "
+                                 "additional selectors to narrow the search.")
+            run_info = result[0].get_info()
+
+        else:
+            # Otherwise create a new node and use that.
+            # We first convert any entity values that are currently collections to
+            # JSON strings to prevent nasty hashing problems downstream. Note that
+            # isinstance() isn't as foolproof as actually trying to hash the
+            # value, but the latter is likely to be slower, and since values are
+            # coming from JSON or filenames, there's no real chance of encountering
+            # anything but a list or dict.
+            entities = {
+                k: (json.dumps(v) if isinstance(v, (list, dict)) else v)
+                for (k, v) in entities.items()
+            }
+
+            run = dataset.create_node('run', entities, image_file=img_f,
+                                      duration=duration, repetition_time=tr)
+            run_info = run.get_info()
 
         # Process event files
         if events:
             dfs = layout.get_nearest(
-                img_f, extension='tsv', suffix='events', all_=True,
+                img_f, extension='.tsv', suffix='events', all_=True,
                 full_search=True, ignore_strict_entities=['suffix', 'extension'])
             for _data in dfs:
                 _data = pd.read_csv(_data, sep='\t')
@@ -180,11 +242,11 @@ def _load_time_variables(layout, dataset=None, columns=None, scan_length=None,
                     if (_data['amplitude'].astype(int) == 1).all() and \
                             'trial_type' in _data.columns:
                         msg = ("Column 'amplitude' with constant value 1 "
-                                "is unnecessary in event files; ignoring it.")
+                               "is unnecessary in event files; ignoring it.")
                         _data = _data.drop('amplitude', axis=1)
                     else:
                         msg = ("Column name 'amplitude' is reserved; "
-                                "renaming it to 'amplitude_'.")
+                               "renaming it to 'amplitude_'.")
                         _data = _data.rename(
                             columns={'amplitude': 'amplitude_'})
                     warnings.warn(msg)
@@ -243,7 +305,7 @@ def _load_time_variables(layout, dataset=None, columns=None, scan_length=None,
 
         if rec_types:
             rec_files = layout.get_nearest(
-                img_f, extension='tsv.gz', all_=True, suffix=rec_types,
+                img_f, extension='.tsv.gz', all_=True, suffix=rec_types,
                 ignore_strict_entities=['suffix', 'extension'], full_search=True)
             for rf in rec_files:
                 metadata = layout.get_metadata(rf)
@@ -294,26 +356,36 @@ def _load_time_variables(layout, dataset=None, columns=None, scan_length=None,
 
 def _load_tsv_variables(layout, suffix, dataset=None, columns=None,
                         prepend_type=False, scope='all', **selectors):
-    ''' Reads variables from scans.tsv, sessions.tsv, and participants.tsv.
+    """Reads variables from scans.tsv, sessions.tsv, and participants.tsv.
 
-    Args:
-        layout (BIDSLayout): The BIDSLayout to use.
-        suffix (str): The suffix of file to read from. Must be one of 'scans',
-            'sessions', or 'participants'.
-        dataset (NodeIndex): A BIDS NodeIndex container. If None, a new one is
-            initialized.
-        columns (list): Optional list of names specifying which columns in the
-            files to return. If None, all columns are returned.
-        prepend_type (bool): If True, variable names are prepended with the
-            type name (e.g., 'age' becomes 'participants.age').
-        scope (str, list): The scope of the space to search for variables. See
-            docstring for BIDSLayout for details and valid predefined values.
-        selectors (dict): Optional keyword arguments passed onto the
-            BIDSLayout instance's get() method; can be used to constrain
-            which data are loaded.
+    Parameters
+    ----------
+    layout : :obj:`bids.layout.BIDSLayout`
+        The BIDSLayout to use.
+    suffix : str
+        The suffix of file to read from. Must be one of 'scans',
+        'sessions', or 'participants'.
+    dataset : NodeIndex
+        A BIDS NodeIndex container. If None, a new one is
+        initialized.
+    columns : list
+        Optional list of names specifying which columns in the
+        files to return. If None, all columns are returned.
+    prepend_type : bool
+        If True, variable names are prepended with the
+        type name (e.g., 'age' becomes 'participants.age').
+    scope : str or list
+        The scope of the space to search for variables. See
+        docstring for BIDSLayout for details and valid predefined values.
+    selectors : dict
+        Optional keyword arguments passed onto the
+        BIDSLayout instance's get() method; can be used to constrain
+        which data are loaded.
 
-    Returns: A NodeIndex instance.
-    '''
+    Returns
+    -------
+    A NodeIndex instance.
+    """
 
     # Sanitize the selectors: only keep entities at current level or above
     remap = {'scans': 'run', 'sessions': 'session', 'participants': 'subject'}
@@ -324,12 +396,12 @@ def _load_tsv_variables(layout, suffix, dataset=None, columns=None,
     if dataset is None:
         dataset = NodeIndex()
 
-    files = layout.get(extension='tsv', suffix=suffix, scope=scope,
+    files = layout.get(extension='.tsv', suffix=suffix, scope=scope,
                        **layout_kwargs)
 
     for f in files:
 
-        _data = pd.read_csv(f.path, sep='\t')
+        _data = f.get_df(include_timing=False)
 
         # Entities can be defined either within the first column of the .tsv
         # file (for entities that vary by row), or from the full file path
@@ -395,6 +467,7 @@ def _load_tsv_variables(layout, suffix, dataset=None, columns=None,
 
         level = {'scans': 'session', 'sessions': 'subject',
                  'participants': 'dataset'}[suffix]
+
         node = dataset.get_or_create_node(level, f.entities)
 
         ent_cols = list(set(ALL_ENTITIES) & set(_data.columns))
